
/****************************************************************************
 *                                                                          *
 *     Numerical Methods subroutines                                        *
 *                                                                          *
 ****************************************************************************/
// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "Constants.h"
#include "Externals.h"
#include "nrutil.h"

#define NRANSI

#define FREERETURN {free_dmatrix(fjac,1,n+1,1,n+1);free_dvector(fvec,1,n+1);\
		      free_dvector(p,1,n+1);free_ivector(indx,1,n+1);return 0;}

int mnewt(int ntrial, int n, double tolx, double tolf, 
	  int &iterations, double *x)
{
  void lubksb(double **a, int n, int *indx, double b[]);
  void ludcmp(double **a, int n, int *indx, double *d);
  extern void usrfun(int n, double *fvec, double **fjac, double *x, 
		     int iterations);
  int k,i,*indx;
  double errx,errf,d,*fvec,**fjac,*p;
  
  indx=ivector(1,n+1);
  p=dvector(1,n+1);
  fvec=dvector(1,n+1);
  fjac=dmatrix(1,n+1,1,n+1);
  for (k=1;k<=ntrial;k++) {
    if (Control.output > 3 && k<=10) {
      double xl=0.0;
      printf("mnewt[%d].x:\n",k);
      for(i=1;i<=n;i++) {
        xl+=x[i]*x[i]; /* norm */
        if (i==6 && 10<=n) printf(" ...");
        if (i<=5 || i>n-6) printf(" % 10.8f",x[i]);
      }
      printf("\nmnewt[%d].x^2: % 10.8e\n",k,xl);
    }
    iterations=k;
    usrfun(n,fvec,fjac,x,iterations);
    errf=0.0;
    for (i=1;i<=n;i++) errf += fabs(fvec[i]);
    if (errf <= tolf) FREERETURN
      for (i=1;i<=n;i++) p[i] = -fvec[i];
    ludcmp(fjac,n,indx,&d);
    lubksb(fjac,n,indx,p);
    errx=0.0;
    for (i=1;i<=n;i++) {
      errx += fabs(p[i]);
      x[i] += p[i];
    }
   if (Control.output > 2)
     printf("mnewt[%d].err=%16.13f\n",k,errx);
   if (errx <= tolx) FREERETURN
  }
  FREERETURN
}
#undef FREERETURN
#undef NRANSI

// in: a[n][n], b[n], indx[n],  out: b[n]
void lubksb(double **a, int n, int *indx, double b[])
{
  int i,ii=0,ip,j;
  double sum;
  
  for (i=1;i<=n;i++) {
    ip=indx[i];
    sum=b[ip];
    b[ip]=b[i];
    if (ii)
      for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
    else if (sum) ii=i;
    b[i]=sum;
  }
  for (i=n;i>=1;i--) {
    sum=b[i];
    for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
    b[i]=sum/a[i][i];
  }
}

#define NRANSI
#define TINY 1.0e-20;

// ToDo: call another faster algorithm for special case n=3? Joerg Sep05
// input:  n = matrix size, a[n][n] 
// output: a[n][n], d={-1,+1}, indx[n]
// ToDo: check taken from NumRec? JS
void ludcmp(double **a, int n, int *indx, double *d)
{
  int i,imax,j,k;
  double big,dum,sum,temp;
  static double *vv=NULL;
  static int oldn=0;
  
  if (vv==NULL || oldn!=n) {
    if (vv) free_dvector(vv,1,oldn+1);
    vv=dvector(1,n+1); oldn=n;  // call it only if n has changed, Joerg Sep05
  }
  *d=1.0;
  for (i=1;i<=n;i++) {
    big=0.0;
    for (j=1;j<=n;j++)
      if ((temp=fabs(a[i][j])) > big) big=temp;
    if (big == 0.0) {
      printf("Singular matrix in routine ludcmp, n=%d i=%d a[n][n]\n", n, i);
      for(i=1;i<=n;i++) for(j=1;j<=n;j++) printf(" %e",a[i][j]);
      exit(1); // was nrerror
    }
    vv[i]=1.0/big;  /* normalize? */
  }
  for (j=1;j<=n;j++) {
    for (i=1;i<j;i++) {
      sum=a[i][j];
      for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
    }
    big=0.0;
    for (imax=i=j;i<=n;i++) { // JS060711 imax= added
      sum=a[i][j];
      for (k=1;k<j;k++) sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
      if ( (dum=vv[i]*fabs(sum)) >= big) {
	big=dum;
	imax=i;
      }
    }
    if (j != imax) {  // exchange column imax with j
      for (k=1;k<=n;k++) {
	dum=a[imax][k]; a[imax][k]=a[j][k]; a[j][k]=dum;
      }
      *d = -(*d);
      vv[imax]=vv[j];
    }
    indx[j]=imax;
    if (a[j][j] == 0.0) a[j][j]=TINY;
    if (j != n) {
      dum=1.0/(a[j][j]);
      for (i=j+1;i<=n;i++) a[i][j] *= dum;
    }
  }
  // free_dvector(vv,1,n+1);
}
#undef TINY
#undef NRANSI

